/* 
 * Sistemas de Tempo Real, 2003/2004
 * Luis Almeida & Paulo Pedreiras
 * Oct.2003
 * 
 * aula3: Linux/RTAI
 * 4 RT tasks toggle a bit on the paralell port.
 * Bit status is sent to rt fifo
 * Linux task reads rt fifo and displays pin status
 *
 * RT module
 *
 * Changelog:
 *  2006/09/21, Paulo Pedreiras
 *     Ported from RTAI 1.0 to RTAI 3.3
 *     Major changes: thread arg in init change dfrom int to long
 *     task_preempt_always() not supported anymore
 *     Makefile updated due to change in linux kernel from 2.4 to 2.6
 *     macro to define init/cleanup functions (kernel 2.4->2.6)
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>
 
/* Application configuration */
#include "app_defs.h"

MODULE_AUTHOR("Luis Almeida and Paulo Pedreiras <{lda,pedreiras}@det.ua.pt>");
MODULE_DESCRIPTION("Runs 4 rt tasks periodically and toggles paralell port bits.");
MODULE_LICENSE("GPL");


#define ONE_SHOT        /* Selects Single Shot timer mode*/
#define DISABLE_PREEMPT /* Uncomment to disable preepmtion */

/* RT task descriptors */
static RT_TASK utask0;
static RT_TASK utask1;
static RT_TASK utask2;
static RT_TASK utask3;

static unsigned char pps=0; /* Paralel port bit image */
static Task_Evt tevt; /* RT task event (time,task_id,event) */

/************************/
/* User Real-Time tasks */
/************************/

static void fun0(long t)
{
  //printk("\n First instance task %d",t);
  while(1) {

#ifdef DISABLE_PREEMPT
    rt_global_cli();
#endif
    
    /* Send start event to fifo */
    //printk("\n %d on",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='s';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    printk("\n %ld on  (%lld)",t,tevt.evt_time);
    
    /* Set paralel port pin */
    pps |= (0x1 << t);
    outb(~pps,LPT_ADDR);
    

    /* Task's work*/
    rt_busy_sleep(TASK0_LOAD);
    

    /*Send finish event to fifo */
    //printk("\n %d off",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='f';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    /* Reset paralel port pin */
    pps &= (~(0x1 << t));
    outb(~pps,LPT_ADDR);

    printk("\n %ld off (%lld)",t,tevt.evt_time);

#ifdef DISABLE_PREEMPT
    rt_global_sti();
#endif
    
    /* Sleep until next activation */
    rt_task_wait_period();
  }
}

static void fun1(long t)
{
  //printk("\n First instance task %d",t);
  while(1) {

#ifdef DISABLE_PREEMPT
    rt_global_cli();
#endif

    /* Send start event to fifo */
    //printk("\n %d on",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='s';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    printk("\n %ld on  (%lld)",t,tevt.evt_time);
    
    /* Set paralel port pin */
    pps |= (0x1 << t);
    outb(~pps,LPT_ADDR);
    

    /* Task's work*/
    rt_busy_sleep(TASK1_LOAD);
    

    /*Send finish event to fifo */
    //printk("\n %d off",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='f';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    /* Reset paralel port pin */
    pps &= (~(0x1 << t));
    outb(~pps,LPT_ADDR);

    printk("\n %ld off (%lld)",t,tevt.evt_time);

#ifdef DISABLE_PREEMPT
    rt_global_sti();
#endif

    /* Sleep until next activation */
    rt_task_wait_period();
  }
}


static void fun2(long t)
{
  //printk("\n First instance task %d",t);
  while(1) {

#ifdef DISABLE_PREEMPT
    rt_global_cli();
#endif

    /* Send start event to fifo */
    //printk("\n %d on",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='s';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    printk("\n %ld on  (%lld)",t,tevt.evt_time);
    
    /* Set paralel port pin */
    pps |= (0x1 << t);
    outb(~pps,LPT_ADDR);
    

    /* Task's work*/
    rt_busy_sleep(TASK2_LOAD);
    

    /*Send finish event to fifo */
    //printk("\n %d off",t);
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='f';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    /* Reset paralel port pin */
    pps &= (~(0x1 << t));
    outb(~pps,LPT_ADDR);

    printk("\n %ld off (%lld)",t,tevt.evt_time);
    
#ifdef DISABLE_PREEMPT
    rt_global_sti();
#endif

    /* Sleep until next activation */
    rt_task_wait_period();
  }
}


static void fun3(long t)
{
  
  //printk("\n First instance task %d",t);
  while(1) {

#ifdef DISABLE_PREEMPT
    rt_global_cli();
#endif

    /* Send start event to fifo */
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='s';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    printk("\n %ld on  (%lld)",t,tevt.evt_time);    

    /* Set paralel port pin */
    pps |= (0x1 << t);
    outb(~pps,LPT_ADDR);
    

    /* Task's work*/
    rt_busy_sleep(TASK3_LOAD);
    

    /*Send finish event to fifo */
    tevt.evt_time=rt_get_time_ns();
    tevt.task_id=t;
    tevt.task_act='f';
    rtf_put(FIFO,&tevt,sizeof(tevt));

    /* Reset paralel port pin */
    pps &= (~(0x1 << t));
    outb(~pps,LPT_ADDR);

    printk("\n %ld off (%lld)",t,tevt.evt_time);
    
#ifdef DISABLE_PREEMPT
    rt_global_sti();
#endif

    /* Sleep until next activation */
    rt_task_wait_period();
  }
}


/***************************************************/
/* Module Initialization and termination functions */
/***************************************************/
int __str_test_init(void)
{
  RTIME tick_period;
  RTIME startTime;

  printk("[str_test] Initializing RTAI module ...\n");

#ifdef DISABLE_PREEMPT
  printk("[str_test] Preemption disabled\n");
#else
  printk("[str_test] Preemption enabled\n");
#endif
  
  /* Create rt fifo */
  rtf_create(FIFO,FIFO_SIZE);

  rt_linux_use_fpu(USE_FPU);

  /* Initialize user RT tasks 
   * args: RT_TASK structure, thread code address, thread  arg (long), stack size, 
   * priority (lower numbers -> higher prio.), FPU use flag, context switch handler
   */
  rt_task_init(&utask0, fun0, 0, STACK_SIZE, 0, USE_FPU, 0);
  rt_task_init(&utask1, fun1, 1, STACK_SIZE, 1, USE_FPU, 0);
  rt_task_init(&utask2, fun2, 2, STACK_SIZE, 2, USE_FPU, 0);
  rt_task_init(&utask3, fun3, 3, STACK_SIZE, 3, USE_FPU, 0);


  /* Set timer mode (one shot vs periodic) */
#ifdef ONE_SHOT
  rt_set_oneshot_mode();
  printk("[str_test] Oneshot mode selected\n");
#else
  printk("[str_test] Periodic timer selected\n");
#endif

  /* Define task preeption: 0-No; 1-Yes */
  //rt_preempt_always(1); 

  /* Compute system tick (tick_period in internal timer units) */
  tick_period = start_rt_timer(nano2count(PERIOD));

  startTime = rt_get_time() + NUM_TASKS*tick_period*1000;
  
  /* Start User RT Tasks */
  rt_task_make_periodic(&utask0, startTime, TASK0_PERIOD*tick_period);
  rt_task_make_periodic(&utask1, startTime, TASK1_PERIOD*tick_period);
  rt_task_make_periodic(&utask2, startTime, TASK2_PERIOD*tick_period);
  rt_task_make_periodic(&utask3, startTime, TASK3_PERIOD*tick_period);
   
  printk("[str_test] Module successfuly loaded!\n");
  return 0;
}


void __str_test_exit(void)
{
    
  stop_rt_timer();
  rt_busy_sleep(10000000);

  /* Remove RT FIFO */
  rtf_destroy(FIFO);

  /* Remove user RT tasks */
  rt_task_delete(&utask0);
  rt_task_delete(&utask1);
  rt_task_delete(&utask2);
  rt_task_delete(&utask3);
  
  printk("\n Finito!\n");
}

module_init(__str_test_init);
module_exit(__str_test_exit);
